﻿using APICycleVDP.DB;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace APICycleVDP.Services
{
    public class VdpCyclesGen(IServiceScopeFactory serviceScopeFactory) : BackgroundService
    {
		private readonly IServiceScopeFactory _serviceScopeFactory = serviceScopeFactory;
		private readonly List<GenCycle> _tasks = [];

		protected override async Task ExecuteAsync(CancellationToken stoppingToken)
        {
			Console.WriteLine("Gen: Start cycles.");

			for (var i = 1; i <= 48; i++)
			{
				var task = new GenCycle(i, _serviceScopeFactory);
				task.Start();
				_tasks.Add(task);
			}

			while (!stoppingToken.IsCancellationRequested)
			{
				try
				{
					await Task.Delay(TimeSpan.FromMinutes(10), stoppingToken);
				}
				catch (TaskCanceledException)
				{
					Console.WriteLine("Gen: Cycle cancelled.");
					continue;
				}
				for (var i = 0; i < _tasks.Count; i++)
				{
					if (i % 10 == 0 && i != 0)
						Console.WriteLine();
					Console.Write(_tasks[i].GetStatus() + "|");
				}
				Console.WriteLine();
			}

			foreach (var task in _tasks)
			{
				task.Stop();
			}
		}
    }

	internal class GenCycle(int vdp, IServiceScopeFactory serviceScopeFactory)
	{
		private readonly int _vdp = vdp;
		private readonly IServiceScopeFactory _serviceScopeFactory = serviceScopeFactory;
		private DbVdpContext? _db;

		private CycleStatus _currCycle = CycleStatus.EndTechCycle;
		private DateTime _factStart = DateTime.Now;
		private DateTime _factEnd = DateTime.Now;
		private DateTime _thinkEnd = DateTime.Now;

		private readonly CancellationTokenSource _cts = new();
		private Task _taskCycle = Task.CompletedTask;

		public void Start()
		{
			_taskCycle = Cycle(_cts.Token);
		}
		public void Stop()
		{
			_cts.Cancel();
			_taskCycle.Wait();
		}

		async public Task Cycle(CancellationToken token)
		{
			using var scope = _serviceScopeFactory.CreateScope();
			_db = scope.ServiceProvider.GetRequiredService<DbVdpContext>();
			while (!token.IsCancellationRequested)
			{
				if (GetCurrCycle()) //Can get info form DB.
				{
					if (DateTime.Now >= _factEnd)
					{
						GetNextCycle();
						GetTimeStart();
						GetTimeThinkEnd();
						GetTimeFactEnd();
						while (!SaveToDB())
						{
							Console.WriteLine("Gen: VDP " + _vdp.ToString("D2") + ": Can't connect to DB.");
							try
							{
								await Task.Delay(5000, token);
							}
							catch (TaskCanceledException)
							{
								Console.WriteLine("Gen: VDP " + _vdp.ToString("D2") + ": Main cycle cancelled.");
								return;
							}
						}
					}
				}
				else //It's new Cycle.
				{
					GetTimeStart();
					GetTimeThinkEnd();
					GetTimeFactEnd();
					while (!SaveToDB())
					{
						Console.WriteLine("Gen: VDP " + _vdp.ToString("D2") + ": Can't connect to DB.");
						try
						{
							await Task.Delay(5000, token);
						}
						catch (TaskCanceledException)
						{
							Console.WriteLine("Gen: VDP " + _vdp.ToString("D2") + ": Main cycle cancelled.");
							return;
						}
					}
				}

				while (!token.IsCancellationRequested && (DateTime.Now < _factEnd))
				{
					var secAwait = (_factEnd - DateTime.Now).TotalSeconds;
					secAwait = secAwait >= 5 ? 5 : secAwait;
					try
					{
						await Task.Delay(Convert.ToInt32(Math.Ceiling(secAwait)) * 1000, token);
					}
					catch (TaskCanceledException)
					{
						Console.WriteLine("Gen: VDP " + _vdp.ToString("D2") + ": Main cycle cancelled.");
						return;
					}
				}
			}
			Console.WriteLine("Gen: VDP " + _vdp.ToString("D2") + ": Main cycle cancelled.");
			return;
		}

		public bool GetCurrCycle()
		{
			if (_db == null)
			{
				return false;
			}
			try
			{
				var tmp = _db.Cycles
					.Where(x => x.NumVdp == _vdp)
					.OrderByDescending(x => x.FactStart)
					.FirstOrDefault();
				if (tmp == null)
				{
					_currCycle = CycleStatus.EndTechCycle;
					return false;
				}
				_currCycle = Enum.IsDefined(typeof(CycleStatus), tmp.NumCycle)
					? (CycleStatus)tmp.NumCycle
					: CycleStatus.EndTechCycle;
				_factStart = tmp.FactStart;
				_factEnd = tmp.FactEnd;
			}
			catch (Exception e)
			{
				Console.WriteLine(e.Message);
				return false;
			}
			return true;
		}
		public void GetNextCycle()
		{
			_currCycle = GenData.GetNextCycle(_currCycle);
		}
		public void GetTimeStart()
		{
			_factStart = _factEnd;
		}
		public void GetTimeThinkEnd()
		{
			_thinkEnd = _factStart.AddMinutes(GenData.GetDuration(_currCycle));
		}
		public void GetTimeFactEnd()
		{
			_factEnd = _thinkEnd.AddSeconds(GenData.GetDeviation(_currCycle));
		}
		public bool SaveToDB()
		{
			if (_db == null)
			{
				return false;
			}
			try
			{
				//Clear DB
				var oldCycles = _db.Cycles
					.Where(x => x.NumVdp == _vdp && x.FactEnd < DateTime.Now.AddMonths(-1));
				if (oldCycles.Any())
				{
					_db.Cycles.RemoveRange(oldCycles);
				}

				//Save parameter
				_db.Cycles.Add(new Cycle()
				{
					NumVdp = _vdp,
					NumCycle = (int)_currCycle,
					FactStart = _factStart,
					FactEnd = _factEnd,
					ThinkEnd = _thinkEnd
				});
				_db.SaveChanges();
			}
			catch (Exception e)
			{
				Console.WriteLine("Gen: " + e.Message);
				return false;
			}
			return true;
		}

		public string GetStatus()
		{
			return _vdp.ToString("D2") + "-" + _currCycle.ToString();
		}
	}

	internal static class GenData
	{
		private static readonly Random random = new();
		private static readonly Dictionary<CycleStatus, (Func<CycleStatus> NextState, int Duration, (int min, int offset))> genDataMap = new()
		{
			{
				CycleStatus.EndTechCycle,
				(() => random.Next(100) < 50
					? CycleStatus.VacForMeltingScarp
					: CycleStatus.VacForWelding,
				15,
				(7 * 60, 5 * 60))
			},
			{
				CycleStatus.LoadUnload,
				(() => CycleStatus.VacForWelding,
				15,
				(3 * 60, 2 * 60))
			},
			{
				CycleStatus.VacForWelding,
				(() => CycleStatus.Welding,
				10,
				(11 * 60, 1 * 60))
			},
			{
				CycleStatus.Welding,
				(() => CycleStatus.CoolingWelding,
				13,
				(4 * 60, 3 * 60))
			},
			{
				CycleStatus.CoolingWelding,
				(() => CycleStatus.CheckWelding,
				7,
				(4 * 60, 3 * 60))
			},
			{
				CycleStatus.CheckWelding,
				(() => random.Next(100) < 20
					? CycleStatus.Welding
					: CycleStatus.VacForMelting,
				5,
				(7 * 60, 2 * 60))
			},
			{
				CycleStatus.VacForMelting,
				(() => CycleStatus.DilutionVat,
				10,
				(11 * 60, 1 * 60))
			},
			{
				CycleStatus.DilutionVat,
				(() => CycleStatus.Melting,
				5,
				(2 * 60, 1 * 60))
			},
			{
				CycleStatus.Melting,
				(() => CycleStatus.BringShrinkageCavity,
				60,
				(40 * 60, 30 * 60))
			},
			{
				CycleStatus.BringShrinkageCavity,
				(() => CycleStatus.CoolingIngot,
				15,
				(5 * 60, 3 * 60))
			},
			{
				CycleStatus.CoolingIngot,
				(() => CycleStatus.EndTechCycle,
				30,
				(20 * 60, 10 * 60))
			},
			{
				CycleStatus.VacForMeltingScarp,
				(() => CycleStatus.MeltingScarp,
				10,
				(11 * 60, 1 * 60))
			},
			{
				CycleStatus.MeltingScarp,
				(() => CycleStatus.CoolingMeltingScarp,
				20,
				(4 * 60, 3 * 60))
			},
			{
				CycleStatus.CoolingMeltingScarp,
				(() => CycleStatus.LoadUnload,
				15,
				(4 * 60, 3 * 60))
			},
		};

		public static CycleStatus GetNextCycle(CycleStatus currCycle)
		{
			CycleStatus nextCycle = CycleStatus.EndTechCycle;
			if (genDataMap.TryGetValue(currCycle, out var data))
			{
				nextCycle = data.NextState();
			}
			return nextCycle;
		}

		public static int GetDuration(CycleStatus currCycle)
		{
			int duration = 15;
			if (genDataMap.TryGetValue(currCycle, out var data))
			{
				duration = data.Duration;
			}
			return duration;
		}

		public static int GetDeviation(CycleStatus currCycle)
		{
			int deviation = random.Next(7 * 60) - (5 * 60);
			if (genDataMap.TryGetValue(currCycle, out var data))
			{
				deviation = random.Next(data.Item3.min) - data.Item3.offset;
			}
			return deviation;
		}
	}

	public enum CycleStatus : int
	{
		EndTechCycle = 0,
		LoadUnload = 1,
		VacForWelding = 2,
		Welding = 5,
		CoolingWelding = 6,
		CheckWelding = 7,
		VacForMelting = 8,
		DilutionVat = 9,
		Melting = 10,
		BringShrinkageCavity = 11,
		CoolingIngot = 12,
		VacForMeltingScarp = 14,
		MeltingScarp = 15,
		CoolingMeltingScarp = 16
	}
}
